"""Accesses Porsche Connect account and retrieves connected vehicles."""

from __future__ import annotations

import logging

from pyporscheconnectapi.connection import Connection
from pyporscheconnectapi.vehicle import PorscheVehicle

_LOGGER = logging.getLogger(__name__)


class PorscheConnectAccount:
    """Establishes a connection to a Porsche Connect account."""

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        token: dict | None = None,
        connection: Connection | None = None,
    ) -> None:
        """Initialize the account."""
        self.vehicles: list[PorscheVehicle] = []
        self.token = token
        if connection is None:
            self.connection = Connection(username, password, token=token)
        else:
            self.connection = connection

    async def _init_vehicles(self) -> None:
        """Initialize vehicles from API endpoint."""
        _LOGGER.debug("Building vehicle list")

        if self.connection is not None:
            vehicle_list = await self.connection.get("/connect/v1/vehicles")

            for vehicle in vehicle_list:
                _LOGGER.debug("Got vehicle %s", vehicle)
                v = PorscheVehicle(
                    vin=vehicle["vin"],
                    data=vehicle,
                    status={},
                    connection=self.connection,
                )
                self.vehicles.append(v)

            self.token = self.connection.token

    async def get_vehicles(self, *, force_init: bool = False) -> list[PorscheVehicle]:
        """Retrieve available vehicles from API endpoints."""
        _LOGGER.debug("Retrieving vehicle list")

        if len(self.vehicles) == 0 or force_init:
            await self._init_vehicles()

        return self.vehicles

    async def get_vehicle(self, vin: str) -> PorscheVehicle | None:
        """Retrieve vehicle data from API endpoints."""
        if len(self.vehicles) == 0:
            await self._init_vehicles()
        filtered = [v for v in self.vehicles if v.vin == vin]
        if len(filtered) > 0:
            return filtered[0]
        return None



#  SPDX-License-Identifier: Apache-2.0
"""Python Package for controlling Porsche Connect API."""

from __future__ import annotations

import asyncio
import logging

import httpx

from .const import API_BASE_URL, TIMEOUT, USER_AGENT, X_CLIENT_ID
from .exceptions import PorscheExceptionError
from .oauth2 import Captcha, Credentials, OAuth2Client, OAuth2Token

_LOGGER = logging.getLogger(__name__)


async def log_request(request):
    """Provide formatting for http logging."""
    _LOGGER.debug("Request headers: %s", request.headers)
    _LOGGER.debug("Request method - url: %s %s", request.method, request.url)
    _LOGGER.debug("Request body: %s", request.content)


class Connection:
    """Handles authentication and connecting to the Porsche Connect API.

    :param email: Porsche Connect email
    :param password: Porsche Connect password
    :param asyncClient: httpx.AsyncClient or None
    :param token: token dict - should be a dict with access_token, refresh_token, expires_at, etc as root params
    :param leeway: time in seconds to consider token as expired before it actually expires
    """

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        captcha_code: str | None = None,
        state: str | None = None,
        async_client=httpx.AsyncClient(),
        token=None,
        leeway: int = 60,
    ) -> None:
        """Initialise the connection to the Porsche Connect API."""
        if token is None:
            token = {}
        self.asyncClient = async_client
        self.token_lock = asyncio.Lock()

        self.token = OAuth2Token(token)

        self.headers = {"User-Agent": USER_AGENT, "X-Client-ID": X_CLIENT_ID}

        self.oauth2_client = OAuth2Client(
            self.asyncClient,
            Credentials(email, password),
            Captcha(captcha_code, state),
            leeway,
        )

    async def get_token(self):
        """Return the authentication token."""
        async with self.token_lock:
            await self.oauth2_client.ensure_valid_token(self.token)
        return self.token

    async def get(self, url, params=None):
        """Make a GET request to the Porsche Connect API."""
        return await self.request("GET", url, params=params)

    async def post(self, url, data=None, json=None):
        """Make a POST request to the Porsche Connect API."""
        return await self.request("POST", url, data=data, json=json)

    async def put(self, url, data=None, json=None):
        """Make a PUT request to the Porsche Connect API."""
        return await self.request("PUT", url, data=data, json=json)

    async def delete(self, url, data=None, json=None):
        """Make a DELETE request to the Porsche Connect API."""
        return await self.request("DELETE", url, data=data, json=json)

    async def request(self, method, url, **kwargs):
        """Create a request to the Porsche Connect API."""
        try:
            async with self.token_lock:
                await self.oauth2_client.ensure_valid_token(self.token)
            resp = await self.asyncClient.request(
                method,
                f"{API_BASE_URL}/{url}",
                headers=self.headers | {"Authorization": f"Bearer {self.token.access_token}"},
                timeout=TIMEOUT,
                **kwargs,
            )
            resp.raise_for_status()  # A common error seem to be: httpx.HTTPStatusError: Server error '504 Gateway Time-out'
            return resp.json()
        except httpx.HTTPStatusError as exc:
            raise PorscheExceptionError(exc.response.status_code) from exc

    async def close(self):
        """Close the asyncClient connection."""
        await self.asyncClient.aclose()



"""Models the state of a Porsche Connect vehicle."""

from __future__ import annotations

import datetime
import json  # only for formatting debug output
import logging
import re
import uuid

from pyporscheconnectapi.connection import Connection
from pyporscheconnectapi.exceptions import PorscheExceptionError
from pyporscheconnectapi.remote_services import RemoteServices

from .const import COMMANDS, MEASUREMENTS, TIRE_PRESSURE_TOLERANCE, TRIP_STATISTICS

_LOGGER = logging.getLogger(__name__)

BASE_DATA = ["vin", "modelName", "customName", "modelType", "systemInfo", "timestamp"]


class PorscheVehicle:
    """Representation of a Porsche Connect vehicle."""

    def __init__(
        self,
        connection: Connection,
        vin: str | None = None,
        data: dict | None = None,
        status: dict | None = None,
        capabilities: dict | None = None,
        trip_statistics: dict | None = None,
        picture_locations: dict | None = None,
    ) -> None:
        """Initialise the Porsche Vehicle."""
        if picture_locations is None:
            picture_locations = {}
        if trip_statistics is None:
            trip_statistics = {}
        if capabilities is None:
            capabilities = {}
        if status is None:
            status = {}
        if data is None:
            data = {}
        if vin is None:
            vin = ""
        self.connection = connection
        self.data = data
        self.remote_services = RemoteServices(self)
        self.status = status
        self.capabilities = capabilities
        self.trip_statistics = trip_statistics
        self.picture_locations = picture_locations

    def get_data(self) -> dict:
        """Get all vechicle data."""
        return self.data

    @property
    def vin(self) -> str:
        """Get the VIN (vehicle identification number) of the vehicle."""
        return self.data.get("vin", "not available")

    @property
    def model_name(self) -> str:
        """Get the model name of the vehicle."""
        return self.data.get("modelName", "not available")

    @property
    def model_year(self) -> str:
        """Get the model year of the vehicle."""
        return self.data.get("modelType", {}).get("year", "not available")

    @property
    def has_porsche_connect(self) -> bool:
        """Should return true if porsche connect subscription is active. Todo."""
        return True

    @property
    def connected(self) -> bool:
        """Returns true if the vechicle is on-line."""
        return self.data.get("connect", "not available")

    @property
    def has_remote_services(self) -> bool:
        """Return true if remote services are available."""
        return self.data.get("REMOTE_ACCESS_AUTHORIZATION", {}).get("isEnabled") is True

    @property
    def has_electric_drivetrain(self) -> bool:
        """Return True if vehicle is equipped with a high voltage battery."""
        return self.data["modelType"]["engine"] == "BEV" or self.data["modelType"]["engine"] == "PHEV"

    @property
    def main_battery_level(self) -> int:
        """Return the main battery level (TODO: for non-BEV it should use the 12V battery level if available?)."""
        level = 0
        if not self.has_ice_drivetrain:
            level = self.data.get("BATTERY_LEVEL", {}).get("percent", 0)
        return level

    @property
    def has_ice_drivetrain(self) -> bool:
        """Return True if vehicle has an internal combustion engine."""
        return self.data["modelType"]["engine"] == "PHEV" or self.data["modelType"]["engine"] == "COMBUSTION"

    @property
    def has_remote_climatisation(self) -> bool:
        """Return True if vehicle has remote climatisation ability."""
        return self.data.__contains__("CLIMATIZER_STATE")

    @property
    def has_direct_charge(self) -> bool:
        """Return True if vehicle has direct charge ability."""
        return self.data.get("BATTERY_CHARGING_STATE", "").__contains__(
            "directChargingState",
        )

    @property
    def direct_charge_on(self) -> bool:
        """Return True if direct charging is enabled."""
        return self.data.get("BATTERY_CHARGING_STATE", {}).get("directChargingState") == "ENABLED_ON"

    @property
    def privacy_mode(self) -> bool:
        """Return True if privacy mode is on."""
        return self.data.get("GLOBAL_PRIVACY_MODE", {}).get("isEnabled")

    @property
    def remote_climatise_on(self) -> bool:
        """Return True if remote climatisation is on."""
        _LOGGER.debug("Remote climatisation is: %s", self.data.get("CLIMATIZER_STATE", {}).get("isOn"))

        return self.data.get("CLIMATIZER_STATE", {}).get("isOn")

    @property
    def vehicle_locked(self) -> bool:
        """Return True if vehicle is locked."""
        return self.data.get("LOCK_STATE_VEHICLE", {}).get("isLocked") is True

    @property
    def vehicle_closed(self) -> bool:
        """Return True if all doors and lids are closed."""
        return not bool(
            sum(
                [self.data[key]["isOpen"] for key in self.data if key.startswith("OPEN_STATE_")],
            ),
        )

    @property
    def doors_and_lids(self) -> bool:
        """Return list of all doors and lids and their status."""
        dl = [{key: "Open" if self.data[key]["isOpen"] is True else "Closed"} for key in self.data if key.startswith("OPEN_STATE_")]
        return dict(map(dict.popitem, dl))

    @property
    def tire_pressure_status(self) -> bool:
        """Return true if tire pressure is within the tolerances."""
        tire_pressure_status = self.data.get("TIRE_PRESSURE")
        return (
            not sorted(
                map(
                    abs,
                    [tire_pressure_status[key]["differenceBar"] for key in tire_pressure_status if key.endswith("Tire")],
                ),
            )[-1]
            > TIRE_PRESSURE_TOLERANCE
        )

    @property
    def tire_pressures(self) -> bool:
        """Return a dict containing tire pressure readings."""
        return self.data.get("TIRE_PRESSURE")

    @property
    def has_tire_pressure_monitoring(self) -> bool:
        """Return True if vehicle has tire pressure monitoring."""
        return self.data.get("TIRE_PRESSURE") is not None

    @property
    def charging_target(self) -> bool | None:
        """Return target state of charge (SoC) for high voltage battery."""
        if self.data.get("CHARGING_PROFILES"):
            charging_profiles = self.data["CHARGING_PROFILES"]["list"]
            active_charging_profile_id = self.data["BATTERY_CHARGING_STATE"].get(
                "activeProfileId",
                None,
            )

            if active_charging_profile_id is None:
                return None

            active_charging_profile = next(
                (p for p in charging_profiles if p["id"] == active_charging_profile_id),
                None,
            )
            _LOGGER.debug("Active charging profile is: %s", active_charging_profile)
            if active_charging_profile is not None:
                return active_charging_profile.get("minSoc")
        return None

    @property
    def location_updated_at(self) -> datetime:
        """Return time stamp of latest location update."""
        datetime_str = self.data.get("GPS_LOCATION", {}).get("lastModified")
        if datetime_str:
            return datetime.datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%SZ").astimezone(datetime.timezone.utc)
        return None

    @property
    def location(self) -> tuple[float | None, float | None, int | None]:
        """Get the location of the vehicle."""
        loc = self.data.get("GPS_LOCATION", {}).get("location")
        heading = self.data.get("GPS_LOCATION", {}).get("direction")
        if loc and re.match(r"[\-\.0-9]+,[\-\.0-9]+", loc):
            lat, lon = map(float, loc.split(","))
        else:
            lat, lon = None, None

        return (lat, lon, heading)

    async def get_stored_overview(self) -> None:
        """Return stored vechicle status overview."""
        measurements = "mf=" + "&mf=".join(MEASUREMENTS)

        try:
            _LOGGER.debug("Getting stored status for vehicle %s", self.vin)
            self.status = await self.connection.get(
                f"/connect/v1/vehicles/{self.vin}?{measurements}",
            )
            self._update_vehicle_data()
        except PorscheExceptionError as err:
            _LOGGER.exception(
                "Could not get stored overview, error communicating with API: '%s",
                err.message,
            )

    async def get_current_overview(self) -> None:
        """Return vehicle current status overview."""
        measurements = "mf=" + "&mf=".join(MEASUREMENTS)
        wakeup = "&wakeUpJob=" + str(uuid.uuid4())

        try:
            _LOGGER.debug("Getting current status for vehicle %s", self.vin)
            self.status = await self.connection.get(
                f"/connect/v1/vehicles/{self.vin}?{measurements+wakeup}",
            )
            self._update_vehicle_data()
        except PorscheExceptionError as err:
            _LOGGER.exception(
                "Could not get current overview, error communicating with API: '%s",
                err.message,
            )

    async def get_capabilities(self) -> None:
        """Return vehicle capabilities."""
        measurements = "mf=" + "&mf=".join(MEASUREMENTS)
        commands = "&cf=" + "&cf=".join(COMMANDS)

        try:
            _LOGGER.debug("Getting capabilities for vehicle %s", self.vin)
            self.capabilities = await self.connection.get(
                f"/connect/v1/vehicles/{self.vin}?{measurements+commands}",
            )
        except PorscheExceptionError as err:
            _LOGGER.exception(
                "Could not get capabilities, error communicating with API: %s",
                err.message,
            )

    async def get_trip_statistics(self) -> None:
        """Return trip statistics for vehicle as a dict."""
        measurements = "mf=" + "&mf=".join(TRIP_STATISTICS)

        try:
            _LOGGER.debug("Getting trip statistics for vehicle %s", self.vin)
            self.trip_statistics = await self.connection.get(
                f"/connect/v1/vehicles/{self.vin}?{measurements}",
            )
        except PorscheExceptionError as err:
            _LOGGER.exception(
                "Could not get capabilities, error communicating with API: %s",
                err.message,
            )

    async def get_picture_locations(self) -> None:
        """Return list of uri's to vechicle pictures."""
        try:
            _LOGGER.debug("Getting picture urls for vehicle %s", self.vin)
            resp = await self.connection.get(
                f"/connect/v1/vehicles/{self.vin}/pictures",
            )
            for p in resp:
                self.picture_locations[p["view"]] = p["url"]
        except PorscheExceptionError as err:
            _LOGGER.exception(
                "Could not get capabilities, error communicating with API: %s",
                err.message,
            )

    def __repr__(self) -> str:
        """Return a printable representation of the Porsche Connect Vehicle object."""
        return f"Vehicle({self.vin!r}, drivetrain={self.data.get('modelType', {}).get('engine')!r}, has_porsche_connect={self.has_porsche_connect!r})"

    def _update_vehicle_data(self) -> None:
        vin = self.vin
        bdata = {}
        mdata = {}

        if "vin" in self.status:
            _LOGGER.debug(
                "Vehicle status dict for %s is now: %s",
                vin,
                json.dumps(self.status, indent=2),
            )

            if "customName" not in self.status:
                self.status["customName"] = ""

            bdata = {k: self.status[k] for k in BASE_DATA}

            bdata["name"] = bdata["customName"] if "customName" in bdata else bdata["modelName"]

            _LOGGER.debug(
                "Got base data for vehicle '%s': %s",
                vin,
                json.dumps(bdata, indent=2),
            )

            if "measurements" in self.status:
                tdata = [m for m in self.status["measurements"] if m["status"]["isEnabled"]]

                for m in tdata:
                    mdata[m["key"]] = m["value"]
                _LOGGER.debug(
                    "Got measurement data for vehicle '%s': %s",
                    vin,
                    json.dumps(mdata, indent=2),
                )

                # Here we do some measurements translations to make them accessible

                if "BATTERY_CHARGING_STATE" in mdata:
                    if "chargingRate" in mdata["BATTERY_CHARGING_STATE"]:
                        # Convert charging rate from km/min to km/h
                        mdata["BATTERY_CHARGING_STATE"]["chargingRate"] = mdata["BATTERY_CHARGING_STATE"]["chargingRate"] * 60
                    else:
                        # Charging is currently not ongoing, but we should still feed some data to the sensor
                        mdata["BATTERY_CHARGING_STATE"]["chargingRate"] = 0

                    if "chargingPower" not in mdata["BATTERY_CHARGING_STATE"]:
                        # Charging is currently not ongoing, but we should still feed some data to the sensor
                        mdata["BATTERY_CHARGING_STATE"]["chargingPower"] = 0

                if "CHARGING_SUMMARY" in mdata:
                    # For some strange reason, the minSoC attribute in this dict does not react on changes,
                    # so we create a shadow of it which we update as required and use that for the sensor instead
                    if mdata["CHARGING_SUMMARY"].get("mode") == "DIRECT":
                        minsoc = 100
                    elif mdata["CHARGING_SUMMARY"].get("chargingProfile"):
                        minsoc = self.charging_target
                    else:
                        _LOGGER.debug("Unable to find minSoC for vehicle '%s", vin)
                        # What should it fall back to?
                        minsoc = 80

                    mdata["CHARGING_SUMMARY"]["minSoC"] = minsoc

            else:
                _LOGGER.debug("Measurement data missing for vehicle '%s", vin)
                _LOGGER.debug(
                    "Payload for current overview query was: %s",
                    json.dumps(self.status, indent=2),
                )

        else:
            _LOGGER.debug("Base data missing for vehicle '%s", vin)
            _LOGGER.debug(
                "Payload for current overview query was: %s",
                json.dumps(self.status, indent=2),
            )

        self.data = self.data | bdata | mdata



"""Library to integrate with the Porsche Connect API.

This library provides a Python interface to Porsche Connect API, with
abilities to read vechicle status data and access to remote services
to control certain vechicle functions.

NOTE: This work is not officially supported by Porsche and functionality
can stop working at any time without warning.

"""



"""Trigger remote services on a vehicle."""

from __future__ import annotations

import asyncio
import datetime
import logging
from enum import Enum
from hashlib import sha512
from typing import TYPE_CHECKING

from .exceptions import PorscheRemoteServiceError

if TYPE_CHECKING:
    from .vehicle import PorscheVehicle
_LOGGER = logging.getLogger(__name__)

#: time in seconds between receiving status and polling for update
_POLLING_DELAY = 1

#: maximum number of seconds to wait for the server to return a positive answer
_POLLING_TIMEOUT = 240


class StrEnum(str, Enum):
    """A string enumeration of type `(str, Enum)`. All members are compared via `upper()`. Defaults to UNKNOWN."""

    @classmethod
    def _missing_(cls, value):
        has_unknown = False
        for member in cls:
            if member.value.upper() == "UNKNOWN":
                has_unknown = True
            if member.value.upper() == value.upper():
                return member
        if has_unknown:
            _LOGGER.warning("'%s' is not a valid '%s'", value, cls.__name__)
            return cls.UNKNOWN
        msg = f"'{value}' is not a valid {cls.__name__}"
        raise ValueError(msg)


class ExecutionState(StrEnum):
    """Enumeration of possible states of the execution of a remote service."""

    PERFORMED = "PERFORMED"
    ERROR = "ERROR"
    UNKNOWN = "UNKNOWN"


class RemoteServiceStatus:
    """Wraps the status of the execution of a remote service."""

    def __init__(self, response: dict, status_id: str | None = None) -> None:
        """Construct a new object from a dict."""
        status = None
        if "status" in response:
            status = response.get("status", {}).get("result")

        self.status = status
        self.state = ExecutionState(status or "UNKNOWN")
        self.details = response
        self.status_id = status_id


class RemoteServices:
    """Trigger remote services on a vehicle."""

    def __init__(self, vehicle: PorscheVehicle):
        """Initialise the Remote Services on a Porsche Connect vehicle."""
        self._vehicle = vehicle
        self._connection = vehicle.connection

    async def flash_indicators(
        self,
    ):
        """Remote service for flashing the indicators briefly."""
        _LOGGER.debug("Requesting vehicle %s to flash indicators.", self._vehicle.vin)

        payload = {
            "key": "HONK_FLASH",
            "payload": {
                "mode": "FLASH",
                "spin": None,
            },
        }

        return await self._send_command(payload)

    async def honk_and_flash_indicators(
        self,
    ):
        """Remote service for honking and flashing the indicators briefly."""
        _LOGGER.debug(
            "Requesting vehicle %s to honk and flash indicators.",
            self._vehicle.vin,
        )

        payload = {
            "key": "HONK_FLASH",
            "payload": {
                "mode": "HONK_AND_FLASH",
                "spin": None,
            },
        }

        return await self._send_command(payload)

    async def climatise_on(
        self,
        *,
        target_temperature: float = 293.15,
        front_left: bool = False,
        front_right: bool = False,
        rear_left: bool = False,
        rear_right: bool = False,
    ):
        """Remote service for turning climatisation on."""
        _LOGGER.debug("Starting remote climatisation for %s", self._vehicle.vin)

        payload = {
            "key": "REMOTE_CLIMATIZER_START",
            "payload": {
                "climateZonesEnabled": {
                    "frontLeft": front_left,
                    "frontRight": front_right,
                    "rearLeft": rear_left,
                    "rearRight": rear_right,
                },
                "targetTemperature": target_temperature,
            },
        }

        return await self._send_command(payload)

    async def climatise_off(
        self,
    ):
        """Remote service for turning climatisation off."""
        _LOGGER.debug("Stopping remote climatisation for %s", self._vehicle.vin)

        payload = {"key": "REMOTE_CLIMATIZER_STOP", "payload": {}}

        return await self._send_command(payload)

    async def direct_charge_on(
        self,
    ):
        """Remote service for turning direct charging on."""
        _LOGGER.debug("Enabling direct charging for %s", self._vehicle.vin)

        payload = {"key": "DIRECT_CHARGING_START", "payload": {"spin": None}}

        return await self._send_command(payload)

    async def direct_charge_off(
        self,
    ):
        """Remote service for turning direct charging off."""
        _LOGGER.debug("Disabling direct charging for %s", self._vehicle.vin)
        payload = {"key": "DIRECT_CHARGING_STOP", "payload": {"spin": None}}
        return await self._send_command(payload)

    async def lock_vehicle(
        self,
    ):
        """Remote service for locking a vehicle."""
        _LOGGER.debug("Locking vehicle %s", self._vehicle.vin)
        payload = {"key": "LOCK", "payload": {"spin": None}}
        return await self._send_command(payload)

    async def unlock_vehicle(
        self,
        pin,
    ):
        """Remote service for unlocking a vehicle."""
        _LOGGER.debug("Unlocking vehicle %s", self._vehicle.vin)

        challenge = await self._get_challenge()

        if challenge:
            pinhash = sha512(bytes.fromhex(pin + challenge)).hexdigest().upper()

            payload = {
                "key": "UNLOCK",
                "payload": {"spin": {"challenge": challenge, "hash": pinhash}},
            }

            return await self._send_command(payload)
        return None

    async def update_charging_profile(
        self,
        profile_id: int | None = None,
        minimum_charge_level: int | None = None,
    ):
        """Remote service for altering a charging profile."""
        chargingprofileslist = self._vehicle.data["CHARGING_PROFILES"]["list"]
        _LOGGER.debug("Charging profile list: %s", chargingprofileslist)

        if profile_id is None:
            profile_id = self._vehicle.data["BATTERY_CHARGING_STATE"]["activeProfileId"]

        _LOGGER.debug("Active profile id: %s", profile_id)

        if minimum_charge_level is not None:
            minimum_charge_level = min(max(int(minimum_charge_level), 25), 100)
            for i, item in enumerate(chargingprofileslist):
                if profile_id == item["id"]:
                    item["minSoc"] = minimum_charge_level
                    chargingprofileslist[i] = item

        return await self._update_charging_profile(chargingprofileslist)

    async def _get_challenge(
        self,
    ):
        payload = {"key": "SPIN_CHALLENGE", "payload": {"spin": None}}

        _LOGGER.debug("Requesting challenge for %s", self._vehicle.vin)

        response = await self._connection.post(
            f"/connect/v1/vehicles/{self._vehicle.vin}/commands",
            json=payload,
        )

        return response.get("data", {}).get("challenge")

    async def _update_charging_profile(
        self,
        chargingprofileslist,
    ):
        payload = {
            "key": "CHARGING_PROFILES_EDIT",
            "payload": {"list": chargingprofileslist},
        }
        _LOGGER.debug("Updating charging profile for %s.", self._vehicle.vin)

        return await self._send_command(payload)

    async def _send_command(
        self,
        payload,
    ):
        _LOGGER.debug("Executing remote command with payload %s", payload)

        response = await self._connection.post(
            f"/connect/v1/vehicles/{self._vehicle.vin}/commands",
            json=payload,
        )

        if response:
            status_id = response.get("status", {}).get("id")
            result_code = response.get("status", {}).get("result")
        else:
            msg = "Did not receive response for remote service request"
            raise PorscheRemoteServiceError(
                msg,
            )

        _LOGGER.debug("Got result: %s (%s)", result_code, status_id)

        status = await self._block_until_done(status_id) if status_id and result_code == "ACCEPTED" else RemoteServiceStatus(result_code)

        await asyncio.sleep(_POLLING_DELAY)
        await self._vehicle.get_stored_overview()

        return status

    async def _block_until_done(self, status_id: str) -> RemoteServiceStatus:
        """Keep polling the server until we get a final answer.

        :raises TimeoutError: if there is no final answer before _POLLING_TIMEOUT
        """
        fail_after = datetime.datetime.now(datetime.UTC) + datetime.timedelta(
            seconds=_POLLING_TIMEOUT,
        )
        status = None
        while datetime.datetime.now(datetime.UTC) < fail_after:
            await asyncio.sleep(_POLLING_DELAY)
            status = await self._get_remote_service_status(status_id)
            _LOGGER.debug("Current state of '%s' is: %s", status_id, status.state.value)
            if status.state == ExecutionState.ERROR:
                msg = f"Remote service failed with state '{status.details}'"
                raise PorscheRemoteServiceError(
                    msg,
                )
            if status.state not in [
                ExecutionState.UNKNOWN,
            ]:
                return status
        current_state = "Unknown"
        if status is not None:
            current_state = status.state.value
        msg = f"Did not receive remote service result for '{status_id}' in {_POLLING_TIMEOUT} seconds. Current state: {current_state}"
        raise PorscheRemoteServiceError(
            msg,
        )

    async def _get_remote_service_status(self, status_id: str) -> RemoteServiceStatus:
        """Return execution status of the last remote service that was triggered."""
        _LOGGER.debug("Getting remote service status for '%s'", status_id)
        status_msg = await self._connection.get(
            f"/connect/v1/vehicles/{self._vehicle.vin}/commands/{status_id}",
        )
        _LOGGER.debug("Got status message %s for '%s'", status_msg, status_id)
        return RemoteServiceStatus(status_msg, status_id=status_id)



"""Client configuration constants."""

AUTHORIZATION_SERVER = "identity.porsche.com"
REDIRECT_URI = "my-porsche-app://auth0/callback"
AUDIENCE = "https://api.porsche.com"
CLIENT_ID = "XhygisuebbrqQ80byOuU5VncxLIm8E6H"
X_CLIENT_ID = "41843fb4-691d-4970-85c7-2673e8ecef40"
USER_AGENT = "pyporscheconnectapi/0.2.0"
API_BASE_URL = "https://api.ppa.porsche.com/app"
AUTHORIZATION_URL = f"https://{AUTHORIZATION_SERVER}/authorize"
TOKEN_URL = f"https://{AUTHORIZATION_SERVER}/oauth/token"
TIMEOUT = 90

SCOPES = [
    "openid",
    "profile",
    "email",
    "offline_access",
    "mbb",
    "ssodb",
    "badge",
    "vin",
    "dealers",
    "cars",
    "charging",
    "manageCharging",
    "plugAndCharge",
    "climatisation",
    "manageClimatisation",
    "pid:user_profile.porscheid:read",
    "pid:user_profile.name:read",
    "pid:user_profile.vehicles:read",
    "pid:user_profile.dealers:read",
    "pid:user_profile.emails:read",
    "pid:user_profile.phones:read",
    "pid:user_profile.addresses:read",
    "pid:user_profile.birthdate:read",
    "pid:user_profile.locale:read",
    "pid:user_profile.legal:read",
]
SCOPE = " ".join(SCOPES)

""" Vehicle properties """

TIRE_PRESSURE_TOLERANCE = 0.3

MEASUREMENTS = [
    "ACV_STATE",
    "ALARM_STATE",
    "BATTERY_CHARGING_STATE",
    "BATTERY_LEVEL",
    "BLEID_DDADATA",
    "CHARGING_PROFILES",
    "CHARGING_SETTINGS",
    "CHARGING_SUMMARY",
    "CLIMATIZER_STATE",
    "DEPARTURES",
    "E_RANGE",
    "FUEL_LEVEL",
    "FUEL_RESERVE",
    "GLOBAL_PRIVACY_MODE",
    "GPS_LOCATION",
    "HEATING_STATE",
    "HVAC_STATE",
    "INTERMEDIATE_SERVICE_RANGE",
    "INTERMEDIATE_SERVICE_TIME",
    "LOCK_STATE_VEHICLE",
    "MAIN_SERVICE_RANGE",
    "MAIN_SERVICE_TIME",
    "MILEAGE",
    "OIL_LEVEL_CURRENT",
    "OIL_LEVEL_MAX",
    "OIL_LEVEL_MIN_WARNING",
    "OIL_SERVICE_RANGE",
    "OIL_SERVICE_TIME",
    "OPEN_STATE_CHARGE_FLAP_LEFT",
    "OPEN_STATE_CHARGE_FLAP_RIGHT",
    "OPEN_STATE_DOOR_FRONT_LEFT",
    "OPEN_STATE_DOOR_FRONT_RIGHT",
    "OPEN_STATE_DOOR_REAR_LEFT",
    "OPEN_STATE_DOOR_REAR_RIGHT",
    "OPEN_STATE_LID_FRONT",
    "OPEN_STATE_LID_REAR",
    "OPEN_STATE_SERVICE_FLAP",
    "OPEN_STATE_SPOILER",
    "OPEN_STATE_SUNROOF",
    "OPEN_STATE_SUNROOF_REAR",
    "OPEN_STATE_TOP",
    "OPEN_STATE_WINDOW_FRONT_LEFT",
    "OPEN_STATE_WINDOW_FRONT_RIGHT",
    "OPEN_STATE_WINDOW_REAR_LEFT",
    "OPEN_STATE_WINDOW_REAR_RIGHT",
    "PAIRING_CODE",
    "PARKING_BRAKE",
    "PARKING_LIGHT",
    "PRED_PRECON_LOCATION_EXCEPTIONS",
    "PRED_PRECON_USER_SETTINGS",
    "RANGE",
    "REMOTE_ACCESS_AUTHORIZATION",
    "SERVICE_PREDICTIONS",
    "THEFT_STATE",
    "TIMERS",
    "TIRE_PRESSURE",
    "VTS_MODES",
]

COMMANDS = [
    "BLEID_AGREEMENT_GIVE",
    "BLEID_AGREEMENT_REVOKE",
    "BLEID_DEVICEKEY_UPLOAD",
    "B_CALL_TRIGGER",
    "CHARGING_PROFILES_EDIT",
    "CHARGING_SETTINGS_AUTOPLUG_EDIT",
    "CHARGING_SETTINGS_BATTERYCAREMODE_EDIT",
    "CHARGING_SETTINGS_CERTIFICATES_RESET",
    "CHARGING_SETTINGS_EDIT",
    "CHARGING_STOP",
    "CS_ACCOUNT_FEDERATION",
    "CS_APP_SHOP_ENABLE",
    "CS_APP_SHOP_UPLOAD",
    "CS_C2P_IN_VEHICLE_INFOTAINMENT",
    "CS_DESTINATION_SYNC",
    "CS_PCM_ACCOUNT_SERVICES",
    "CS_PCM_CALENDAR",
    "CS_PILOTED_PARKING",
    "CS_VOICE_MIMIC",
    "CS_YOUNITED",
    "CS_VIDEOSTREAMING_VOUCHER",
    "DEPARTURES_EDIT",
    "DIRECT_CHARGING_START",
    "DIRECT_CHARGING_STOP",
    "HONK_FLASH",
    "LOCK",
    "PRED_PRECON_LOCATION_EXCEPTION_EDIT",
    "PRED_PRECON_USER_SETTINGS_EDIT",
    "REMOTE_ACV_START",
    "REMOTE_ACV_STOP",
    "REMOTE_CLIMATIZER_START",
    "REMOTE_CLIMATIZER_STOP",
    "REMOTE_HEATING_START",
    "REMOTE_HEATING_STOP",
    "ROUTE_CALCULATE",
    "SERVICE_PREDICTIONS_VISIBILITY_EDIT",
    "SPIN_CHALLENGE",
    "SPIN_VALIDATION",
    "TIMERS_DISABLE",
    "TIMERS_EDIT",
    "UNLOCK",
]

TRIP_STATISTICS = [
    "TRIP_STATISTICS_CYCLIC",
    "TRIP_STATISTICS_LONG_TERM",
    "TRIP_STATISTICS_LONG_TERM_HISTORY",
    "TRIP_STATISTICS_SHORT_TERM_HISTORY",
    "TRIP_STATISTICS_CYCLIC_HISTORY",
    "TRIP_STATISTICS_SHORT_TERM",
]



"""Authentication token management for Porsche Connect API."""

#  SPDX-License-Identifier: Apache-2.0
import asyncio
import logging
import time
from typing import NamedTuple
from urllib.parse import parse_qs, urlparse

import httpx
from bs4 import BeautifulSoup

from .const import (
    AUDIENCE,
    AUTHORIZATION_SERVER,
    AUTHORIZATION_URL,
    CLIENT_ID,
    REDIRECT_URI,
    SCOPE,
    TIMEOUT,
    TOKEN_URL,
    USER_AGENT,
    X_CLIENT_ID,
)
from .exceptions import (
    PorscheCaptchaRequiredError,
    PorscheExceptionError,
    PorscheWrongCredentialsError,
)

_LOGGER = logging.getLogger(__name__)


class Credentials(NamedTuple):
    """Store credentials for the Porsche Connect API."""

    email: str
    password: str


class Captcha(NamedTuple):
    """Store captcha data for the Porsche Connect API."""

    captcha_code: str
    state: str


class OAuth2Token(dict):
    """A simple wrapper around a dict to handle OAuth2 tokens.

    Provides a helper method to check if the token is expired.
    Originally based on: https://github.com/lepture/authlib/blob/master/authlib/oauth2/rfc6749/wrappers.py
    """

    def __init__(self, params: dict):
        """Initialise the oauth2 token."""
        if params.get("expires_at"):
            self["expires_at"] = int(params["expires_at"])
        elif params.get("expires_in"):
            self.expires_at = params["expires_in"]
        super().__init__(params)

    def is_expired(self, leeway=60):
        """Return true if the access token has expired."""
        expires_at = self.get("expires_at")
        if not expires_at:
            return None
        # small timedelta to consider token as expired before it actually expires
        expiration_threshold = expires_at - leeway
        return expiration_threshold < time.time()

    @property
    def expires_at(self):
        """Return the expiration time stamp of the access token."""
        return self.get("expires_at")

    @property
    def access_token(self):
        """Return the access token."""
        return self.get("access_token")

    @property
    def refresh_token(self):
        """Return the refresh token."""
        return self.get("refresh_token")

    @expires_at.setter
    def expires_at(self, expires_in):
        self["expires_at"] = int(time.time()) + int(expires_in)


class OAuth2Client:
    """Utility class to handle OAuth2 authentication with Porsche Connect.

    :param client: httpx.AsyncClient
    :param credentials: tuple of email, password
    :param leeway: time in seconds to consider token as expired before it actually expires
    """

    def __init__(
        self,
        client: httpx.AsyncClient,
        credentials: Credentials,
        captcha: Captcha,
        leeway: int = 60,
    ):
        """Initialise the oauth2 client."""
        self.client = client
        self.credentials = credentials
        self.captcha = captcha
        self.leeway = leeway
        self.headers = {"User-Agent": USER_AGENT, "X-Client-ID": X_CLIENT_ID}

    async def ensure_valid_token(self, token: OAuth2Token):
        """Ensure the access_token is valid, logging in or refreshing if necessary."""
        token_is_expired = token.is_expired(self.leeway)
        if token_is_expired:
            token_data = await self.refresh_token(token.refresh_token)
            token.update(token_data)
            token.expires_at = token_data["expires_in"]
            _LOGGER.debug("Refreshed Access Token: %s", token.access_token)
        if token.access_token is None or token_is_expired is None:  # no token, get a new one
            auth_code = await self.fetch_authorization_code()
            token_data = await self.fetch_access_token(auth_code)
            token.update(token_data)
            token.expires_at = token_data["expires_in"]
            _LOGGER.debug("New Access Token: %s", token.access_token)

    async def fetch_authorization_code(self):
        """Fetch the authorization code from Porsche Connect.

        Requires 1-4 requests (1 if already logged in, 4 if not):

        1. Initial request to /authorize to get the code
        2. If no code is returned, login with Identifier First flow:
            2a. POST to /u/login/identifier with email
            2b. POST to /u/login/password with password
        3. Resume the /authorize request with the resume path from the Identifier First flow

        :return: authorization code to be exchanged for an access token
        """
        if self.captcha.captcha_code is None:
            try:
                _LOGGER.debug("Fetching authorization code.")

                # first request to get the code
                params = await self.get_and_extract_location_params(
                    AUTHORIZATION_URL,
                    params={
                        "response_type": "code",
                        "client_id": CLIENT_ID,
                        "redirect_uri": REDIRECT_URI,
                        "audience": AUDIENCE,
                        "scope": SCOPE,
                        "state": "pyporscheconnectapi",
                    },
                )
                authorization_code = params.get("code", [None])[0]

                # if we already have a session with Auth, just use the code they return
                if authorization_code is not None:
                    _LOGGER.debug("Got authorization code: %s", authorization_code)
                    return authorization_code

                # no existing Auth0 session, run through Identifier First flow
                _LOGGER.debug(
                    "No existing auth0 session, running through identifier first flow.",
                )

                resume_path = await self.login_with_identifier(params["state"][0])

                # completed the Identifier First flow, now resume the auth code request
                params = await self.get_and_extract_location_params(
                    f"https://{AUTHORIZATION_SERVER}{resume_path}",
                )
                authorization_code = params.get("code", [None])[0]

            except httpx.HTTPStatusError as exc:
                raise PorscheExceptionError(exc.response.status_code) from exc

            else:
                _LOGGER.debug("Authorization code: %s", authorization_code)
                return authorization_code

        else:
            try:
                resume_path = await self.login_with_identifier(self.captcha.state)
                params = await self.get_and_extract_location_params(
                    f"https://{AUTHORIZATION_SERVER}{resume_path}",
                )
                authorization_code = params.get("code", [None])[0]

            except httpx.HTTPStatusError as exc:
                raise PorscheExceptionError(exc.response.status_code) from exc

            else:
                _LOGGER.debug("Authorization code: %s", authorization_code)
                return authorization_code

    async def get_and_extract_location_params(self, url, params=None):
        """GET the URL and extract the params from the Location header.

        :param url: URL to GET
        :param params: dict of query parameters
        :return: dict of query parameters from the Location header
        """
        if params is None:
            params = {}
        resp = await self.client.get(
            url,
            params=self._merge_query_params(url, params),
            timeout=TIMEOUT,
            headers=self.headers,
        )
        if resp.status_code != 302:
            msg = "Could not fetch authorization code"
            raise PorscheExceptionError(msg)

        location = resp.headers["Location"]
        return self._extract_params_from_url(location)

    def _extract_params_from_url(self, url):
        """Extract the query parameters from a URL.

        :param url: URL to extract the query parameters from
        :return: dict of query parameters
        """
        return parse_qs(urlparse(url).query)

    def _merge_query_params(self, url: str, params: dict[str, str]) -> dict[str, str]:
        """Merge query parameters into a new dictionary with the existing query parameters of a URL."""
        parsed_url = urlparse(url)
        query = parse_qs(parsed_url.query)
        new_query = {k: v[0] for k, v in query.items()}
        new_query.update(params)
        return new_query

    async def login_with_identifier(self, state: str):
        """Log into the Identifier First flow.

        Takes 2 steps:

        1. POST to /u/login/identifier with email
        2. POST to /u/login/password with password

        :param state: state parameter from the initial authorize request
        :return: URL to resume the auth code request
        """
        # 1. /u/login/identifier w/ email (and captcha code)

        data = {
            "state": state,
            "username": self.credentials.email,
            "js-available": True,
            "webauthn-available": False,
            "is-brave": False,
            "webauthn-platform-available": False,
            "action": "default",
        }

        if self.captcha.captcha_code is None:
            _LOGGER.debug("Submitting e-mail address to auth endpoint.")
        else:
            data.update({"captcha": self.captcha.captcha_code})
            _LOGGER.debug(
                "Submitting e-mail address and captcha code %s to auth endpoint.",
                self.captcha.captcha_code,
            )

        url = f"https://{AUTHORIZATION_SERVER}/u/login/identifier"
        resp = await self.client.post(
            url,
            data=data,
            params={"state": state},
            timeout=TIMEOUT,
            headers=self.headers,
        )

        if resp.status_code == 401:
            msg = "Wrong credentials"
            raise PorscheWrongCredentialsError(msg)

        # In case captcha verification is required, the response code is 400 and the captcha is provided as a svg image
        if resp.status_code == 400:
            _LOGGER.debug("Captcha required.")
            soup = BeautifulSoup(resp.text, "html.parser")
            captcha_img = soup.find("img", {"alt": "captcha"})["src"]
            _LOGGER.debug("Parsed out SVG captcha: %s", captcha_img)
            raise PorscheCaptchaRequiredError(captcha=captcha_img, state=state)

        # 2. /u/login/password w/ password

        _LOGGER.debug("Submitting password to auth endpoint.")

        data = {
            "state": state,
            "username": self.credentials.email,
            "password": self.credentials.password,
            "action": "default",
        }

        url = f"https://{AUTHORIZATION_SERVER}/u/login/password"
        resp = await self.client.post(
            url,
            data=data,
            params={"state": state},
            timeout=TIMEOUT,
            headers=self.headers,
        )

        # In case of wrong password, the response code is 400 (Bad request)
        if resp.status_code == 400:
            _LOGGER.debug("Invalid credentials.")
            msg = "Wrong credentials"
            raise PorscheWrongCredentialsError(msg)

        resume_url = resp.headers["Location"]
        _LOGGER.debug("Resume at %s:", resume_url)

        _LOGGER.debug("Sleeping 2.5s...")
        await asyncio.sleep(2.5)

        return resume_url

    async def fetch_access_token(self, authorization_code):
        """Exchanges the authorization code for an access token.

        :param authorization_code: authorization code from the /authorize request
        :return: access token
        """
        data = {
            "client_id": CLIENT_ID,
            "grant_type": "authorization_code",
            "code": authorization_code,
            "redirect_uri": REDIRECT_URI,
        }

        try:
            _LOGGER.debug("Exchanging the authorization code for an access token.")

            resp = await self.client.post(
                TOKEN_URL,
                data=data,
                timeout=TIMEOUT,
                headers=self.headers,
            )
            resp.raise_for_status()
            return resp.json()
        except httpx.HTTPStatusError as exc:
            raise PorscheExceptionError(exc.response.status_code) from exc

    async def refresh_token(self, refresh_token):
        """Use the provided refresh token to get a new access token.

        :param refresh_token: refresh token
        :return: access token
        """
        data = {
            "client_id": CLIENT_ID,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }
        try:
            _LOGGER.debug("Using the refresh token to get a new access token.")

            resp = await self.client.post(
                TOKEN_URL,
                data=data,
                timeout=TIMEOUT,
                headers=self.headers,
            )
            resp.raise_for_status()
            return resp.json()
        except httpx.HTTPStatusError as exc:
            # 403 usually means the refresh token is invalid
            # clear the access token so the full login flow can happen again
            if exc.response.status_code == 403:
                return {"access_token": None, "expires_in": 0}
            raise PorscheExceptionError(exc.response.status_code) from exc



#!/usr/bin/python

"""Command line interface for Porsche Connect API functions."""

import argparse
import asyncio
import configparser
import json
import logging
import sys
from getpass import getpass
from pathlib import Path

from pyporscheconnectapi.account import PorscheConnectAccount
from pyporscheconnectapi.connection import Connection
from pyporscheconnectapi.exceptions import PorscheWrongCredentialsError
from pyporscheconnectapi.remote_services import RemoteServices

vehicle_commands = {
    "battery": "Prints the main battery level (BEV)",
    "capabilities": "Get vehicle capabilities",
    "chargingprofile": "Update parameters in configured charging profile",
    "climatise_off": "Stop remote climatisation",
    "climatise_on": "Start remote climatisation",
    "connected": "Check if vehicle is on-line",
    "currentoverview": "Get stored overview for vehicle",
    "direct_charge_off": "Disable direct charging",
    "direct_charge_on": "Enable direct charging",
    "doors_and_lids": "List status of all doors and lids",
    "flash_indicators": "Flash indicators",
    "honk_and_flash": "Flash indicators and sound the horn",
    "location": "Show location of vehicle",
    "lock_vehicle": "Lock vehicle",
    "pictures": "Get vehicle pictures url",
    "storedoverview": "Poll vehicle for current overview",
    "tire_status": "Check if tire pressure are ok",
    "tire_pressures": "Get tire pressure readings",
    "trip_statistics": "Get trip statistics from backend",
    "unlock_vehicle": "Unlock vehicle",
    "vehicle_closed": "Check if all doors and lids are closed",
}

try:
    from rich.console import Console

    console = Console()
    printc = console.print
except ImportError:
    pass

logging.basicConfig()
logging.root.setLevel(logging.WARNING)

_LOGGER = logging.getLogger(__name__)


async def battery(vehicle, _args):
    """Get vehicle battery state of charge (%)."""
    await vehicle.get_stored_overview()
    return vehicle.main_battery_level


async def capabilities(vehicle, _args):
    """Get vehicle capabilities."""
    await vehicle.get_capabilities()
    return vehicle.capabilities


async def chargingprofile(vehicle, args):
    """Manipulate charging profile."""
    await vehicle.get_stored_overview()
    service = RemoteServices(vehicle)
    result = await service.update_charging_profile(
        profile_id=args.profileid,
        minimum_charge_level=args.minimumchargelevel,
    )
    return result.status


async def climatise_off(vehicle, _args):
    """Stop climatisation."""
    service = RemoteServices(vehicle)
    result = await service.climatise_off()
    return result.status


async def climatise_on(vehicle, _args):
    """Start climatisation."""
    service = RemoteServices(vehicle)
    result = await service.climatise_on()
    return result.status


async def connected(vehicle, _args):
    """Get vehicle on-line status."""
    await vehicle.get_current_overview()
    return vehicle.connected


async def currentoverview(vehicle, _args):
    """Get current overview from vehicle."""
    await vehicle.get_current_overview()
    return vehicle.data


async def direct_charge_off(vehicle, _args):
    """Disable direct charging."""
    service = RemoteServices(vehicle)
    result = await service.direct_charge_off()
    return result.status


async def direct_charge_on(vehicle, _args):
    """Enable direct charging."""
    service = RemoteServices(vehicle)
    result = await service.direct_charge_on()
    return result.status


async def doors_and_lids(vehicle, _args):
    """Get stored status of doors and lids from backend."""
    await vehicle.get_stored_overview()
    return vehicle.doors_and_lids


async def flash_indicators(vehicle, _args):
    """Flash the indicators briefly."""
    service = RemoteServices(vehicle)
    result = await service.flash_indicators()
    return result.status


async def honk_and_flash(vehicle, _args):
    """Honk and flash the indicators briefly."""
    service = RemoteServices(vehicle)
    result = await service.honk_and_flash_indicators()
    return result.status


async def location(vehicle, _args):
    """Get the location of the vehicle."""
    await vehicle.get_stored_overview()
    return vehicle.location


async def lock_vehicle(vehicle, _args):
    """Lock the vehicle."""
    service = RemoteServices(vehicle)
    result = await service.lock_vehicle()
    return result.status


async def pictures(vehicle, _args):
    """Get pictures (uri) of the vehicle."""
    await vehicle.get_picture_locations()
    return vehicle.picture_locations


async def storedoverview(vehicle, _args):
    """Get stored overview from back-end."""
    await vehicle.get_stored_overview()
    return vehicle.data


async def tire_status(vehicle, _args):
    """Get tire pressure status from back-end."""
    await vehicle.get_stored_overview()
    return vehicle.tire_pressure_status


async def tire_pressures(vehicle, _args):
    """Get tire pressures from back-end."""
    await vehicle.get_stored_overview()
    return vehicle.tire_pressures


async def trip_statistics(vehicle, _args):
    """Get pictures (uri) of the vehicle."""
    await vehicle.get_trip_statistics()
    return vehicle.trip_statistics


async def unlock_vehicle(vehicle, args):
    """Unock the vehicle."""
    service = RemoteServices(vehicle)
    result = await service.unlock_vehicle(args.pin)
    return result.status


async def vehicle_closed(vehicle, _args):
    """Check with backend if doors and lids are closed."""
    await vehicle.get_stored_overview()
    return vehicle.vehicle_closed


async def main(args):
    """Get arguments from parser and run command."""
    try:
        with Path.open(args.session_file) as json_file:
            token = json.load(json_file)
    except FileNotFoundError:
        token = {}
    except json.decoder.JSONDecodeError:
        token = {}

    if args.debug:
        logging.root.setLevel(logging.DEBUG)

    email = args.email or input("Please enter Porsche Connect email: ")
    password = args.password or getpass()

    connection = Connection(email, password, token=token)
    controller = PorscheConnectAccount(connection=connection)

    response = {}
    try:
        if args.command == "list":
            vehicles = await controller.get_vehicles()
            for vehicle in vehicles:
                response = response | vehicle.data
        elif args.command == "token":
            response = controller.token
        elif args.vin is not None:
            vins = [args.vin]
        elif args.all:
            vehicles = await controller.get_vehicles()
            vins = (v.vin for v in vehicles)
        else:
            sys.exit("--vin or --all is required")
        for vin in vins:
            vehicle = await controller.get_vehicle(vin)
            if vehicle is not None:
                response = await globals()[args.func](vehicle, args)
    except PorscheWrongCredentialsError as e:
        sys.exit(e.message)
    else:
        printc(response)
    await connection.close()
    with Path.open(args.session_file, "w", encoding="utf-8") as json_file:
        json.dump(connection.token, json_file, ensure_ascii=False, indent=2)


def add_arg_vin(parser):
    """Add vin to the argument parser."""
    group = parser.add_mutually_exclusive_group(
        required=True,
    )
    group.add_argument("-v", "--vin", dest="vin", default=None)
    group.add_argument("-a", "--all", dest="all", action="store_true")


def cli():
    """Get configuration parameters and command line argumentsn and run main loop."""
    config = configparser.ConfigParser()
    config["porsche"] = {
        "email": "",
        "password": "",
        "session_file": ".session",
    }
    config.read([".porscheconnect.cfg", Path("~/.porscheconnect.cfg").expanduser()])
    parser = argparse.ArgumentParser(description="Porsche Connect CLI")
    subparsers = parser.add_subparsers(help="command help", dest="command")

    parser.add_argument("-d", "--debug", dest="debug", action="store_true")
    parser.add_argument(
        "-e",
        "--email",
        dest="email",
        default=config.get("porsche", "email"),
    )
    parser.add_argument(
        "-p",
        "--password",
        dest="password",
        default=config.get("porsche", "password"),
    )
    parser.add_argument(
        "-s",
        "--sessionfile",
        dest="session_file",
        default=config.get("porsche", "session_file"),
    )

    subparsers.add_parser("list")
    subparsers.add_parser("token")

    for vcmd, vdesc in vehicle_commands.items():
        parser_command = subparsers.add_parser(vcmd, help=vdesc)
        parser_command.set_defaults(func=vcmd)
        add_arg_vin(parser_command)
        if vcmd == "unlock_vehicle":
            parser_command.add_argument(
                "-n",
                "--pin",
                required=True,
                dest="pin",
                default=None,
            )
        if vcmd == "chargingprofile":
            parser_command.add_argument(
                "--profileid",
                dest="profileid",
                type=int,
                required=False,
                help="Profile id",
            )
            parser_command.add_argument(
                "--chargelevel",
                dest="minimumchargelevel",
                type=int,
                required=False,
                default=None,
                help="Minimum charge level",
            )
            parser_command.add_argument(
                "--profileactive",
                dest="profileactive",
                type=bool,
                required=False,
                default=None,
                help="Profile active status",
            )

    args = parser.parse_args()

    if args.command:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(main(args))
    else:
        parser.print_help(sys.stderr)
